import json
import zipfile
import logging

from .http import *
from .exceptions import *


logger = logging.getLogger("skafos")


def upload_version(files, description=None, **kwargs) -> dict:
    """
    Upload a model version (one or more files) for a specific app and model to Skafos. All files
    are automatically zipped together and uploaded to storage. Once successfully uploaded, a dictionary
    of meta data is returned.

    .. note:: If your model file(s) are not in your working directory, Skafos will zip up and preserve the entire
              path pointing to the file(s). We recommend placing your file(s) in your current working directory.

    :param files:
        Single model file path or list of file paths to zip up and upload to Skafos.
    :type files:
        str or list
    :param description:
        *Optional*. Short description for your model version. Must be less than or equal to 255 characters.
    :type description:
        str
    :param \**kwargs:
        Keyword arguments identifying the organization, app, and model for upload. See below.

    :Keyword Args:
        * *skafos_api_token* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_API_TOKEN`.
        * *org_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_ORG_NAME`.
        * *app_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_APP_NAME`.
        * *model_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_MODEL_NAME`.

    :returns:
            A meta data dictionary for the uploaded model version.

    :Usage:
    .. sourcecode:: python

       from skafos import models

       # Upload a model version to Skafos
       models.upload_version(
           skafos_api_token="<your-api-token>",
           org_name="my-organization",
           app_name="my-app",
           model_name="my-model",
           files="my_text_classifier.mlmodel",
           description="My model version upload to Skafos"
       )
    """
    # Get required params
    params = generate_required_params(kwargs)

    # Unzip files
    filelist = None
    if isinstance(files, list):
        filelist = files
    elif isinstance(files, str):
        filelist = [files]

    # Create zipped filename
    zip_name = params["model_name"] if params["model_name"].endswith(".zip") else "{}.zip".format(params['model_name'])
    body = {"filename": zip_name}

    # Check that they
    if description and isinstance(description, str):
        if len(description) > 255:
            raise InvalidParamError("Description too long. Please provide a description that is less than 255 characters")
        else:
            body["description"] = description
    if description and not isinstance(description, str):
        raise InvalidParamError("Please provide a model version description with type = str")

    # Write the zip file
    with zipfile.ZipFile(zip_name, "w", zipfile.ZIP_DEFLATED) as skazip:
        for zfile in filelist:
            if os.path.isdir(zfile):
                for root, dirs, files in os.walk(zfile):
                    for dfile in files:
                        skazip.write(os.path.join(root, dfile))
            elif os.path.isfile(zfile):
                skazip.write(zfile)
            else:
                raise InvalidParamError("We were unable to find {}. Check to make sure that the file path is correct.".format(zfile))

    # Create endpoint
    endpoint = "/organizations/{org_name}/apps/{app_name}/models/{model_name}/".format(**params)

    # Create a model version
    model_version_res = http_request(
        method="POST",
        url=API_BASE_URL + endpoint + "model_versions",
        payload=json.dumps(body),
        api_token=params["skafos_api_token"]
    ).json()

    # Upload the model
    if model_version_res.get("presigned_url"):
        with open(zip_name, "rb") as data:
            model_data = data.read()
        logger.info("Uploading model version to Skafos")
        upload_res = http_request(
            method="PUT",
            url=model_version_res["presigned_url"],
            header={"Content-Type": "application/octet-stream"},
            payload=model_data,
            api_token=params["skafos_api_token"]
        )
    else:
        raise UploadFailedError("Upload failed.")

    # Update the model version with the file path
    if upload_res.status_code == 200:
        model_version_endpoint = endpoint + "model_versions/{model_version_id}".format(**model_version_res)
        data = {"filepath": model_version_res["filepath"]}
        final_model_version_res = http_request(
            method="PATCH",
            url=API_BASE_URL + model_version_endpoint,
            payload=json.dumps(data),
            api_token=params["skafos_api_token"]
        ).json()
    else:
        raise UploadFailedError("Upload failed.")

    # Return cleaned response JSON to the user
    logger.info("\nSuccessful Upload\n")
    meta = {k: final_model_version_res[k] for k in final_model_version_res.keys() & {"version", "description", "name", "model"}}
    return meta


def fetch_version(version=None, **kwargs):
    """
    Download a model version as a zipped archive for a specific app and model from Skafos to your current
    working directory as `<model_name>.zip`.

    :param version:
        Version of the model to download. If unspecified, defaults to the latest version.
    :type version:
        int
    :param \**kwargs:
        Keyword arguments identifying the organization, app, and model for download. See below.

    :Keyword Args:
        * *skafos_api_token* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_API_TOKEN`.
        * *org_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_ORG_NAME`.
        * *app_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_APP_NAME`.
        * *model_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_MODEL_NAME`.

    :return:
        None

    :Usage:
    .. sourcecode:: python

       from skafos import models

       # Fetch the latest version of your model
       models.fetch_version(
           skafos_api_token="<your-api-token>",
           org_name="my-organization",
           app_name="my-app",
           model_name="my-model"
       )

       # Fetch a specific version of your model
       models.fetch_version(
           skafos_api_token="<your-api-token>",
           org_name="my-organization",
           app_name="my-app",
           model_name="my-model",
           version=2
       )
    """

    # Get required params
    params = generate_required_params(kwargs)

    # Check to warn about overwriting file
    if os.path.exists(params["model_name"] + ".zip"):
        raise DownloadFailedError("A zip file already exists in your working directory with the model name you're trying to download.")

    # Get model version and create endpoint.
    endpoint = "/organizations/{org_name}/apps/{app_name}/models/{model_name}".format(**params)
    if version:
        if isinstance(version, int):
            endpoint += "?version={}".format(version)
        else:  # You passed in a non-supported version
            raise InvalidParamError("If specified, the model version must be an integer.")

    # Download the model
    logger.info("Fetching model")
    method = "GET"
    res = http_request(
        method=method,
        url=DOWNLOAD_BASE_URL + endpoint,
        api_token=params["skafos_api_token"],
        stream=True
    )

    # Log message
    if res.status_code == 200:
        logger.info("Downloaded as {}".format(params["model_name"] + ".zip"))


def list_versions(**kwargs) -> list:
    """
    Return a list of all saved model versions based on organization, app name, and model name.

    :param \**kwargs:
        Keyword arguments identifying the organization, app, and model for version retrieval. See below.

    :Keyword Args:
        * *skafos_api_token* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_API_TOKEN`.
        * *org_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_ORG_NAME`.
        * *app_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_APP_NAME`.
        * *model_name* (``str``) --
            If not provided, it will be read from the environment as `SKAFOS_MODEL_NAME`.

    :return:
        List of dictionaries containing model versions that have been successfully uploaded to Skafos.
    :rtype:
        list

    :Usage:
    .. sourcecode:: python

       from skafos import models

       # List previously saved model versions
       models.list_versions(
           skafos_api_token="<your-api-token>",
           org_name="my-organization",
           app_name="my-app",
           model_name="my-model"
       )
    """
    params = generate_required_params(kwargs)
    endpoint = "/organizations/{org_name}/apps/{app_name}/models/{model_name}/model_versions?order_by=version".format(**params)
    res = http_request(
        method="GET",
        url=API_BASE_URL + endpoint,
        api_token=params["skafos_api_token"]
    ).json()
    # Clean up the response so users have something manageable
    versions = []
    for model_version in res:
        versions.append({k: model_version[k] for k in model_version.keys() & {"version", "name", "updated_at", "description"}})
    return versions



import os
import requests
import logging

from .exceptions import *


API_BASE_URL = "https://api.skafos.ai/v2"
DOWNLOAD_BASE_URL = "https://download.skafos.ai/v2"
HTTP_VERBS = ["GET", "POST", "PUT", "PATCH"]
DEFAULT_TIMEOUT = 120
logger = logging.getLogger(name="skafos")


def generate_required_params(args):
    # Generate the parameters to build a Skafos request/endpoint
    params = {}

    # Check for api token first
    if 'skafos_api_token' in args:
        params['skafos_api_token'] = args['skafos_api_token']
    else:
        params['skafos_api_token'] = os.getenv('SKAFOS_API_TOKEN')
    if not params['skafos_api_token']:
        raise InvalidTokenError("Missing Skafos API Token")

    # Grab org name
    if 'org_name' in args:
        params['org_name'] = args['org_name']
    else:
        params['org_name'] = os.getenv('SKAFOS_ORG_NAME')
    if not params['org_name']:
        raise InvalidParamError("Missing Skafos Organization Name")

    # Grab app name
    if 'app_name' in args:
        params['app_name'] = args['app_name']
    else:
        params['app_name'] = os.getenv('SKAFOS_APP_NAME')
    if not params['app_name']:
        raise InvalidParamError("Missing Skafos App Name")

    # Grab model name
    if 'model_name' in args:
        params['model_name'] = args['model_name']
    else:
        params['model_name'] = os.getenv('SKAFOS_MODEL_NAME')
    if not params['model_name']:
        raise InvalidParamError("Missing Skafos Model Name")

    return params


def http_request(method, url, api_token, header=None, timeout=None, payload=None, stream=False):
    # Check that we ae using an appropriate request type
    if method not in HTTP_VERBS:
        raise requests.exceptions.HTTPError("Must use an appropriate HTTP verb")

    # Prepare headers and timeout
    request_header = {"X-API-TOKEN": api_token, "Content-Type": "application/json"}
    # Update request header with optionally provided header
    if header and isinstance(header, dict):
        request_header.update(header)
    if not timeout:
        timeout = DEFAULT_TIMEOUT

    # Prepare request object and send it
    try:
        req = requests.Request(method, url, headers=request_header, data=payload)
        r = req.prepare()
        with requests.Session() as s:
            logger.debug("Sending prepared request with url: {}".format(url))
            if stream and method == "GET":
                with s.send(r, timeout=timeout, stream=True) as response:
                    response.raise_for_status()
                    fn = url.split("models/")[1].split("?")[0]
                    with open(fn + ".zip", 'wb') as f:
                        for chunk in response.iter_content(chunk_size=512*1024):
                            if chunk:
                                f.write(chunk)
            else:
                response = s.send(r, timeout=timeout)
                response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        logger.debug("HTTP Error: {}".format(err))
        if response.status_code == 401:
            # We know what it is - raise proper exception to user
            raise InvalidTokenError("Invalid Skafos API Token")
        elif response.status_code == 404:
            # We know what it is - raise proper exception to user
            raise InvalidParamError("Invalid param passed to function. Check your org name, app name, or model name")
        else:
            raise
    except requests.exceptions.ConnectionError as err:
        logger.debug("Error connecting to server: {}".format(err))
        raise
    except requests.exceptions.Timeout:
        logger.debug("Request timed out at {} seconds, consider increasing timeout".format(timeout))
        raise
    except requests.exceptions.RequestException as err:
        logger.debug("Oops, got some other error: {}".format(err))
        raise

    # Return response
    logger.debug("Got a 200 from the server")
    return response




class InvalidParamError(Exception):
    """Exception thrown when there is some missing or invalid parameter
    required for one of the SDK utilities."""
    pass


class InvalidTokenError(Exception):
    """Exception thrown when there is either a missing or
    invalid API token when a request is made to Skafos."""
    pass


class UploadFailedError(Exception):
    """Exception thrown when there is an error while uploading
    a model version to Skafos."""
    pass


class DownloadFailedError(Exception):
    """Exception thrown when there is an error while downloading
    a model version from Skafos."""
    pass



import os
from .http import http_request, API_BASE_URL
from .exceptions import InvalidTokenError


def get_version():
    """Returns the current version of the Skafos SDK in use.

    :Usage:
    .. sourcecode:: python

       import skafos

       skafos.get_version()

    """
    with open(os.path.join(os.path.dirname(__file__),  "VERSION")) as version_file:
        return version_file.read().strip()


def summary(skafos_api_token=None) -> dict:
    """
    Returns all Skafos organizations, apps, and models that the provided API token has access to.

    :param skafos_api_token:
        Skafos API Token associated with the user account. Checks environment for 'SKAFOS_API_TOKEN' if not passed
        into the function directly. Get one at https://dashboard.skafos.ai --> Account Settings --> Tokens.
    :type skafos_api_token:
        str or None
    :return:
        Nested dictionary of all organizations, apps, and models this user has access to.
    :rtype:
        dict

    :Usage:
    .. sourcecode:: python

       import skafos

       skafos.summary(skafos_api_token="<YOUR-SKAFOS-API-TOKEN>")
    """
    summary_res = {}

    # Check for api token first
    if not skafos_api_token:
        skafos_api_token = os.getenv("SKAFOS_API_TOKEN")
    if not skafos_api_token:
        raise InvalidTokenError("Missing Skafos API Token")

    # Prepare requests
    method = "GET"
    endpoint = "/organizations"
    res = http_request(
        method=method,
        url=API_BASE_URL + endpoint,
        api_token=skafos_api_token
    ).json()
    for org in res:
        summary_res[org["display_name"]] = {}
        endpoint = "/organizations/{}/apps?with_models=true".format(org["display_name"])
        res = http_request(
            method=method,
            url=API_BASE_URL + endpoint,
            api_token=skafos_api_token
        ).json()
        for app in res:
            summary_res[org["display_name"]][app["name"]] = []
            for model in app["models"]:
                model_meta_data = {k: model[k] for k in model.keys() & {"name", "updated_at"}}
                summary_res[org["display_name"]][app["name"]].append(model_meta_data)

    # Return the summary response to the user
    return summary_res



import os
import setuptools


def read(fname):
  """Utility function to read the README and VERSION files."""
  return open(os.path.join(os.path.dirname(__file__), fname)).read()


with open('requirements.txt', 'r') as r:
    REQS = r.read().splitlines()


setuptools.setup(
  name="skafos",
  packages=setuptools.find_packages(),
  version=read('skafos/VERSION'),
  license="Apache Software License",
  description="Python wrapper for loading, fetching, and listing model versions with the Skafos platform.",
  long_description=read('README.md'),
  long_description_content_type="text/markdown",
  author="Skafos, LLC",
  author_email="skafos@skafos.ai",
  url="https://github.com/skafos/skafossdk",
  download_url='',
  keywords=['machine learning delivery', 'mobile deployment', 'model versioning'],
  install_requires=REQS,
  include_package_data=True,
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3'
  ],
  project_urls={
    "Dashboard": "https://dashboard.skafos.ai",
    "SDK Documentation": "https://sdk.skafos.ai",
    "Source": "https://github.com/skafos/skafossdk",
    "Website": "https://skafos.ai"
  },
  python_requires=">=3"
)
